# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from gettext import gettext as _
import os

from django.conf import settings
from django.db import transaction
from drf_yasg.utils import swagger_auto_schema
from rest_framework.decorators import detail_route
from rest_framework import serializers, status
from rest_framework.response import Response

from pulpcore.plugin.models import Artifact
from pulpcore.plugin.serializers import (
    AsyncOperationResponseSerializer,
    RepositoryPublishURLSerializer,
    RepositorySyncURLSerializer,
)

from pulpcore.plugin.tasking import enqueue_with_reservation

from pulpcore.plugin.viewsets import (
    ContentFilter,
    ContentViewSet,
    RemoteViewSet,
    OperationPostponedResponse,
    PublisherViewSet,
)

from . import tasks
from .models import CookbookPackageContent, CookbookRemote, CookbookPublisher
from .serializers import (
    CookbookPackageContentSerializer,
    CookbookRemoteSerializer,
    CookbookPublisherSerializer,
)

from pulp_cookbook.metadata import CookbookMetadata


class CookbookPackageContentFilter(ContentFilter):
    """Filters for the content endpoint."""

    class Meta:
        model = CookbookPackageContent
        fields = ["name", "version", "content_id"]


class CookbookPackageContentViewSet(ContentViewSet):
    """The ViewSet for the content endpoint."""

    endpoint_name = "cookbooks"
    queryset = CookbookPackageContent.objects.all()
    serializer_class = CookbookPackageContentSerializer
    filterset_class = CookbookPackageContentFilter

    @transaction.atomic
    def create(self, request):
        data = request.data
        try:
            artifact = self.get_resource(data["_artifact"], Artifact)
        except KeyError:
            raise serializers.ValidationError(detail={"_artifact": _("This field is required")})

        abs_filepath = os.path.join(settings.MEDIA_ROOT, artifact.file.name)
        try:
            metadata = CookbookMetadata.from_cookbook_file(abs_filepath, data["name"])
        except KeyError:
            raise serializers.ValidationError(detail={"name": _("This field is required")})
        except FileNotFoundError:
            raise serializers.ValidationError(
                detail={"_artifact": _("No metadata.json found in cookbook tar")}
            )

        try:
            if data["version"] != metadata.version:
                raise serializers.ValidationError(
                    detail={"version": _("version does not correspond to version in cookbook tar")}
                )
        except KeyError:
            pass
        data["version"] = metadata.version
        serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        serializer.save(
            dependencies=metadata.dependencies,
            content_id_type=CookbookPackageContent.SHA256,
            content_id=artifact.sha256,
        )
        headers = self.get_success_headers(request.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)


class CookbookRemoteViewSet(RemoteViewSet):
    """The ViewSet for the remote endpoint."""

    endpoint_name = "cookbook"
    queryset = CookbookRemote.objects.all()
    serializer_class = CookbookRemoteSerializer

    @swagger_auto_schema(
        operation_description="Trigger an asynchronous task to sync cookbook " "content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @detail_route(methods=("post",), serializer_class=RepositorySyncURLSerializer)
    def sync(self, request, pk):
        """
        Synchronizes a repository. The ``repository`` field has to be provided.
        """
        remote = self.get_object()
        serializer = RepositorySyncURLSerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)
        repository = serializer.validated_data.get("repository")
        mirror = serializer.validated_data["mirror"]
        result = enqueue_with_reservation(
            tasks.synchronize,
            [repository, remote],
            kwargs={"remote_pk": remote.pk, "repository_pk": repository.pk, "mirror": mirror},
        )
        return OperationPostponedResponse(result, request)


class CookbookPublisherViewSet(PublisherViewSet):
    """The ViewSet for the publish endpoint."""

    endpoint_name = "cookbook"
    queryset = CookbookPublisher.objects.all()
    serializer_class = CookbookPublisherSerializer

    @swagger_auto_schema(
        operation_description="Trigger an asynchronous task to publish " "cookbook content.",
        responses={202: AsyncOperationResponseSerializer},
    )
    @detail_route(methods=("post",), serializer_class=RepositoryPublishURLSerializer)
    def publish(self, request, pk):
        """
        Publishes a repository.

        Either the ``repository`` or the ``repository_version`` fields can
        be provided but not both at the same time.

        """
        publisher = self.get_object()
        serializer = RepositoryPublishURLSerializer(data=request.data, context={"request": request})
        serializer.is_valid(raise_exception=True)
        repository_version = serializer.validated_data.get("repository_version")

        result = enqueue_with_reservation(
            tasks.publish,
            [repository_version.repository, publisher],
            kwargs={
                "publisher_pk": str(publisher.pk),
                "repository_version_pk": str(repository_version.pk),
            },
        )
        return OperationPostponedResponse(result, request)



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from unittest.mock import patch

from django.test import TestCase

from pulpcore.plugin.models import Artifact, ContentArtifact, RemoteArtifact
from pulpcore.plugin.content import Handler

from pulp_cookbook.app.models import CookbookPackageContent, CookbookRemote
from pulp_cookbook.app.content import CookbookContentHandler


class CookbookContentHandlerTestCase(TestCase):
    """Test the CookbookContentHandler."""

    def setUp(self):
        self.remote = CookbookRemote.objects.create(name="remote")

        self.c1 = CookbookPackageContent.objects.create(name="c1", version="1.0.0", dependencies={})
        self.ca1 = ContentArtifact.objects.create(
            artifact=None, content=self.c1, relative_path=self.c1.relative_path()
        )
        self.ra1 = RemoteArtifact.objects.create(content_artifact=self.ca1, remote=self.remote)
        self.c1_prime = CookbookPackageContent.objects.create(
            name="c1", version="1.0.0", dependencies={}
        )
        self.ca1_prime = ContentArtifact.objects.create(
            artifact=None, content=self.c1_prime, relative_path=self.c1_prime.relative_path()
        )
        self.ra1_prime = RemoteArtifact.objects.create(
            content_artifact=self.ca1_prime, remote=self.remote
        )
        self.assertEqual(self.c1.content_id_type, CookbookPackageContent.UUID)
        self.assertEqual(self.c1_prime.content_id_type, CookbookPackageContent.UUID)

    @patch.object(Handler, "_save_artifact", return_value=Artifact(sha256="1"))
    def test_save_artifact(self, save_artifact_mock):
        """Verify the 'on_demand' policy case."""
        cch = CookbookContentHandler()
        new_artifact = cch._save_artifact(None, RemoteArtifact.objects.get(pk=self.ra1.pk))
        c1 = CookbookPackageContent.objects.get(pk=self.c1.pk)
        self.assertIsNotNone(new_artifact)
        self.assertEqual(c1.content_id, "1")

    @patch.object(Handler, "_save_artifact", return_value=Artifact(sha256="1"))
    def test_save_artifact_other_content_exists(self, save_artifact_mock):
        """When save tries to 'upgrade' the cookbook with the digest, failure is ignored."""
        CookbookPackageContent.objects.create(
            name="c1",
            version="1.0.0",
            content_id_type=CookbookPackageContent.SHA256,
            content_id="1",
            dependencies={},
        )

        cch = CookbookContentHandler()
        new_artifact = cch._save_artifact(None, RemoteArtifact.objects.get(pk=self.ra1.pk))
        c1 = CookbookPackageContent.objects.get(pk=self.c1.pk)
        self.assertIsNotNone(new_artifact)
        self.assertEqual(c1.content_id_type, CookbookPackageContent.UUID)
        self.assertEqual(c1.content_id, str(self.c1.content_id))



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from gettext import gettext as _

from rest_framework import serializers

from pulpcore.plugin.serializers import (
    SingleArtifactContentSerializer,
    RemoteSerializer,
    PublisherSerializer,
)

from .models import CookbookPackageContent, CookbookRemote, CookbookPublisher


class CookbookPackageContentSerializer(SingleArtifactContentSerializer):
    """Serializer for the cookbook content."""

    name = serializers.CharField(help_text=_("name of the cookbook"))
    version = serializers.CharField(help_text=_("version of the cookbook"), required=False)
    dependencies = serializers.JSONField(
        help_text=_("dependencies of the cookbook"), read_only=True
    )
    content_id_type = serializers.HiddenField(default=CookbookPackageContent.SHA256)
    content_id = serializers.CharField(
        help_text=_(
            "content_id of the cookbook (UUID (lazy download)/SHA256 (immediate download/import)"
        ),
        read_only=True,
    )

    def validate(self, data):
        """Validate the CookbookPackageContent data."""
        data = super().validate(data)
        data["_relative_path"] = CookbookPackageContent.relative_path_from_data(data)
        return data

    def update(self, instance, validated_data):
        raise serializers.ValidationError("content is immutable")

    class Meta:
        fields = tuple(set(SingleArtifactContentSerializer.Meta.fields) - {"_relative_path"}) + (
            "name",
            "version",
            "dependencies",
            "content_id_type",
            "content_id",
        )
        model = CookbookPackageContent


class CookbookRemoteSerializer(RemoteSerializer):
    """Serializer for the remote pointing to a universe repo."""

    cookbooks = serializers.JSONField(
        help_text=_(
            'An optional JSON object in the format {"<cookbook name>":'
            ' "<version_string>" }. Used to limit the cookbooks to synchronize'
            " from the remote"
        ),
        required=False,
    )

    class Meta:
        fields = RemoteSerializer.Meta.fields + ("cookbooks",)
        model = CookbookRemote

    # TODO: No support for version specifiers yet, only cookbook names
    def validate_cookbooks(self, value):
        if value == "":  # blank value
            return value
        if isinstance(value, dict):
            if all(value.keys()):
                return value
        raise serializers.ValidationError(
            _('Format must be {"<cookbook_name>" : "version_string" }')
        )


class CookbookPublisherSerializer(PublisherSerializer):
    """Serializer for the publisher."""

    class Meta:
        fields = PublisherSerializer.Meta.fields
        model = CookbookPublisher



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from aiohttp import web

from .handler import CookbookContentHandler, pulp_cookbook_content_path

from pulpcore.plugin.content import app


app.add_routes(
    [
        web.get(
            pulp_cookbook_content_path() + "{path:.+}/universe",
            CookbookContentHandler().handle_universe,
        )
    ]
)
app.add_routes(
    [web.get(pulp_cookbook_content_path() + "{path:.+}", CookbookContentHandler().stream_content)]
)



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Tests that sync cookbook plugin repositories."""
import unittest

from pulp_smash import api, config
from pulp_smash.exceptions import TaskReportError
from pulp_smash.pulp3.constants import DOWNLOAD_POLICIES as ALL_DOWNLOAD_POLICIES
from pulp_smash.pulp3.constants import REPO_PATH
from pulp_smash.pulp3.utils import delete_orphans, gen_remote, gen_repo, sync

from pulp_cookbook.tests.functional.constants import (
    fixture_u1,
    fixture_u1_diff_digest,
    COOKBOOK_REMOTE_PATH,
    DOWNLOAD_POLICIES,
)
from pulp_cookbook.tests.functional.api.utils import (
    get_cookbook_added_content,
    get_cookbook_content,
    get_cookbook_removed_content,
)


class SyncCookbookRepoTestCase(unittest.TestCase):
    """Sync repositories with the cookbook plugin."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()

    def setUp(self):
        delete_orphans(self.cfg)

    def verify_counts(self, repo, all_count, added_count, removed_count):
        self.assertEqual(len(get_cookbook_content(repo)), all_count)
        self.assertEqual(len(get_cookbook_added_content(repo)), added_count)
        self.assertEqual(len(get_cookbook_removed_content(repo)), removed_count)

    def verify_ids_and_artifacts(self, repo, policy):
        """
        Verify that content_id and artifact presence corresponds to the policy.

        Verify that cookbook content_id is of the right type (based in the
        format; content_type_id is hidden by the serializer). Artifacts must be
        present in 'immediate' mode, otherwise not.
        """
        for cookbook in get_cookbook_content(repo):
            if policy == "immediate":
                self.assertEqual(
                    len(cookbook["content_id"]),
                    64,
                    msg=f"{cookbook} does not have a SHA256 content_id",
                )
                self.assertIsNotNone(cookbook["_artifact"])
            else:
                self.assertEqual(
                    len(cookbook["content_id"]),
                    36,
                    msg=f"{cookbook} does not have a UUID conten_id",
                )
                self.assertIsNone(cookbook["_artifact"])

    def sync_and_inspect_task_report(
        self, remote, repo, download_count, mirror=None, policy="immediate"
    ):
        """Do a sync and verify the number of downloaded artifacts.

        Returns:
            Task report structure

        """
        if mirror is None:
            sync_resp = sync(self.cfg, remote, repo)
        else:
            sync_resp = sync(self.cfg, remote, repo, mirror=mirror)
        tasks = tuple(api.poll_spawned_tasks(self.cfg, sync_resp))
        self.assertEqual(len(tasks), 1)
        for report in tasks[0]["progress_reports"]:
            if report["message"] == "Downloading Artifacts":
                if policy != "immediate":
                    self.fail(f"'Downloading Artifacts' stage in task report for {policy} policy")
                self.assertEqual(report["done"], download_count)
                break
        else:
            if policy == "immediate":
                self.fail("Could not find 'Downloading Artifacts' stage in task report")
        return tasks[0]

    def do_create_repo_and_sync(self, client, policy):
        """
        Create a repo and remote (fixture_u1) using `policy`. Sync the repo.

        Verify that a new version was created, the number of downloads and the
        number of content units.
        """
        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["_href"])

        body = gen_remote(fixture_u1.url, policy=policy)
        remote = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["_href"])

        # Sync the full repository.
        self.assertIsNone(repo["_latest_version_href"])

        all_cookbook_count = fixture_u1.cookbook_count()
        task = self.sync_and_inspect_task_report(remote, repo, all_cookbook_count, policy=policy)

        repo = client.get(repo["_href"])

        latest_version_href = repo["_latest_version_href"]
        self.assertIsNotNone(latest_version_href)
        self.assertEqual(latest_version_href, task["created_resources"][0])
        self.verify_counts(repo, all_cookbook_count, all_cookbook_count, 0)

        return repo, remote

    def do_create_repo_and_sync_twice(self, client, policy, second_policy):
        """
        Create a repo and remote (fixture_u1). Sync the repo twice.

        The first sync happens with policy `policy`, the second sync with policy
        `second_policy`.

        Verify that a new version was created, the number of downloads and the
        number of content units for both syncs.
        """
        all_cookbook_count = fixture_u1.cookbook_count()

        repo, remote = self.do_create_repo_and_sync(client, policy)
        latest_version_href = repo["_latest_version_href"]

        body = gen_remote(fixture_u1.url, policy=second_policy)
        client.patch(remote["_href"], body)

        # Sync the full repository again using the second policy.
        exp_download_count = 0
        exp_policy = second_policy
        if policy != "immediate" and second_policy == "immediate":
            exp_download_count = all_cookbook_count
        if policy == "immediate":
            exp_policy = "immediate"
        self.sync_and_inspect_task_report(
            remote, repo, exp_download_count, policy=second_policy, mirror=True
        )
        repo = client.get(repo["_href"])
        self.assertNotEqual(latest_version_href, repo["_latest_version_href"])
        # When we download the actual artifacts, the respective content unit will be replaced.
        # This looks like adding/deleting all cookbooks.
        self.verify_counts(repo, all_cookbook_count, exp_download_count, exp_download_count)
        self.verify_ids_and_artifacts(repo, exp_policy)

        return repo, remote

    def do_sync_check(self, policy):
        """Sync repository with the cookbook plugin.

        In order to sync a repository a remote has to be associated within
        this repository. When a repository is created this version field is set
        as None. After a sync the repository version is updated.

        Do the following:

        1.  Delete orphan content units, create a repository, and an remote.
        2.  Assert that repository version is None
        3.  Sync the remote.
        4.  Assert that repository version is not None and the artifact/content counts
        5.  Add a filter for a single cookbook name to the remote.
        6.  Sync the remote.
        7.  Assert artifact/content counts
        8. Change the filter for a single cookbook name in the remote.
        9. Sync the remote with "mirror=False".
        10. Assert artifact/content counts (content must have been added)
        """
        client = api.Client(self.cfg, api.json_handler)

        repo, remote = self.do_create_repo_and_sync(client, policy)

        # Sync the repository with a filter (use mirror mode).
        all_cookbook_count = fixture_u1.cookbook_count()
        latest_version_href = repo["_latest_version_href"]

        client.patch(remote["_href"], {"cookbooks": {fixture_u1.example1_name: ""}})
        self.sync_and_inspect_task_report(remote, repo, 0, policy=policy, mirror=True)
        repo = client.get(repo["_href"])
        self.assertNotEqual(latest_version_href, repo["_latest_version_href"])
        example1_count = fixture_u1.cookbook_count([fixture_u1.example1_name])
        self.verify_counts(repo, example1_count, 0, all_cookbook_count - example1_count)

        # Sync the repository with another filter and add cookbooks (additive mode (default).
        client.patch(remote["_href"], {"cookbooks": {fixture_u1.example2_name: ""}})
        # Although cookbook content is already present, it is not present in the
        # repository. Thus, it must be downloaded again.
        example2_count = fixture_u1.cookbook_count([fixture_u1.example2_name])
        self.sync_and_inspect_task_report(remote, repo, example2_count, policy=policy)
        repo = client.get(repo["_href"])
        self.assertNotEqual(latest_version_href, repo["_latest_version_href"])
        self.verify_counts(
            repo,
            fixture_u1.cookbook_count([fixture_u1.example1_name, fixture_u1.example2_name]),
            fixture_u1.cookbook_count([fixture_u1.example2_name]),
            0,
        )

        # Verify that cookbook content_id is of the right type (content_type_id
        # is hidden by the serializer)
        self.verify_ids_and_artifacts(repo, policy)

    def test_sync_immediate(self):
        self.do_sync_check("immediate")

    def test_sync_on_demand(self):
        self.do_sync_check("on_demand")

    def test_sync_streamed(self):
        self.do_sync_check("streamed")

    def test_sync_immediate_immediate(self):
        client = api.Client(self.cfg, api.json_handler)
        self.do_create_repo_and_sync_twice(client, "immediate", "immediate")

    def test_sync_on_demand_immediate(self):
        client = api.Client(self.cfg, api.json_handler)
        self.do_create_repo_and_sync_twice(client, "on_demand", "immediate")

    def test_sync_immediate_on_demand(self):
        client = api.Client(self.cfg, api.json_handler)
        self.do_create_repo_and_sync_twice(client, "immediate", "on_demand")

    def test_sync_on_demand_on_demand(self):
        client = api.Client(self.cfg, api.json_handler)
        self.do_create_repo_and_sync_twice(client, "on_demand", "on_demand")

    def test_sync_repo_isolation(self):
        """Sync two repositories with same content but different artifacts.

        The two repo fixtures u1 and u1_diff_digest contain the same cookbooks
        by name and version, but differ in the digest of the artifact. Ensure
        that syncing these two remotes to two repos respectively does not mixup
        cookbooks.
        """
        client = api.Client(self.cfg, api.json_handler)

        # Create repo u1 and sync partially
        repo_u1 = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo_u1["_href"])

        body = gen_remote(fixture_u1.url, cookbooks={fixture_u1.example1_name: ""})
        remote_u1 = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote_u1["_href"])

        self.assertIsNone(repo_u1["_latest_version_href"])

        example1_count = fixture_u1.cookbook_count([fixture_u1.example1_name])
        self.sync_and_inspect_task_report(remote_u1, repo_u1, example1_count)

        repo_u1 = client.get(repo_u1["_href"])
        self.verify_counts(
            repo_u1, all_count=example1_count, added_count=example1_count, removed_count=0
        )

        # Create repo u1_diff_digest and do a full sync
        repo_u1_diff_digest = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo_u1_diff_digest["_href"])

        body = gen_remote(fixture_u1_diff_digest.url)
        remote_u1_diff_digest = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote_u1_diff_digest["_href"])

        self.assertIsNone(repo_u1_diff_digest["_latest_version_href"])

        # u1 and u1_diff_digest must not share content: all cookbooks are added
        cookbook_count = fixture_u1_diff_digest.cookbook_count()
        self.sync_and_inspect_task_report(
            remote_u1_diff_digest, repo_u1_diff_digest, cookbook_count
        )

        repo_u1_diff_digest = client.get(repo_u1_diff_digest["_href"])
        self.verify_counts(
            repo_u1_diff_digest,
            all_count=cookbook_count,
            added_count=cookbook_count,
            removed_count=0,
        )

        # Full sync u1
        client.patch(remote_u1["_href"], {"cookbooks": {}})
        all_cookbook_count = fixture_u1.cookbook_count()
        self.sync_and_inspect_task_report(remote_u1, repo_u1, all_cookbook_count - example1_count)
        repo_u1 = client.get(repo_u1["_href"])
        self.verify_counts(
            repo_u1,
            all_count=all_cookbook_count,
            added_count=all_cookbook_count - example1_count,
            removed_count=0,
        )

        # Verify that the same cookbooks (w.r.t. name and version) differ by content_id
        content_u1_diff_digest = get_cookbook_content(repo_u1_diff_digest)
        self.assertTrue(content_u1_diff_digest)
        for c_u1 in get_cookbook_content(repo_u1):
            for c_u1_diff_digest in content_u1_diff_digest:
                if (
                    c_u1["name"] == c_u1_diff_digest["name"]
                    and c_u1["version"] == c_u1_diff_digest["version"]
                ):
                    artifact_u1 = client.get(c_u1["_artifact"])
                    self.assertEqual(c_u1["content_id"], artifact_u1["sha256"])
                    artifact_u1_diff_digest = client.get(c_u1_diff_digest["_artifact"])
                    self.assertEqual(
                        c_u1_diff_digest["content_id"], artifact_u1_diff_digest["sha256"]
                    )
                    self.assertNotEqual(c_u1["content_id"], c_u1_diff_digest["content_id"])
                    break
            else:
                self.fail(f"Found no matching cookbook for {c_u1} in u1_diff_digest")


class SyncInvalidTestCase(unittest.TestCase):
    """Sync a repository with an invalid given url on the remote."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.cfg = config.get_config()
        cls.client = api.Client(cls.cfg, api.json_handler)

    def test_invalid_url(self):
        """Sync a repository using a remote url that does not exist.

        Test that we get a task failure. See :meth:`do_test`.
        """
        context = self.do_test("http://example.com/invalid/")
        self.assertIsNotNone(context.exception.task["error"]["description"])

    def test_invalid_policies(self):
        """Sync a repository using all unsupported policies.

        Test that we get a task failure when using an unsupported policy. See
        :meth:`do_test`.
        """
        for policy in ALL_DOWNLOAD_POLICIES:
            if policy not in DOWNLOAD_POLICIES:
                context = self.do_test(fixture_u1.url, policy=policy)
                self.assertIsNotNone(context.exception.task["error"]["description"])

    def do_test(self, url, **remote_kwargs):
        """Sync a repository given ``url`` on the remote."""
        repo = self.client.post(REPO_PATH, gen_repo())
        self.addCleanup(self.client.delete, repo["_href"])
        body = gen_remote(url=url, **remote_kwargs)
        remote = self.client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(self.client.delete, remote["_href"])
        with self.assertRaises(TaskReportError) as context:
            sync(self.cfg, remote, repo, mirror=True)
        return context



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Utilities for cookbook plugin tests."""
import functools

from pulp_smash import utils
from pulp_smash.pulp3.utils import get_added_content, get_content, get_removed_content

from pulp_cookbook.tests.functional.constants import COOKBOOK_CONTENT_NAME


def gen_publisher(**kwargs):
    """Return a semi-random dict for use in creating a publisher."""
    data = {"name": utils.uuid4()}
    data.update(kwargs)
    return data


def _filter_for_cookbook_content(get_func, *args, **kwargs):
    content = get_func(*args, **kwargs)
    assert not content or list(content.keys()) == [COOKBOOK_CONTENT_NAME]
    return content.get(COOKBOOK_CONTENT_NAME, {})


get_cookbook_content = functools.partial(_filter_for_cookbook_content, get_content)
get_cookbook_added_content = functools.partial(_filter_for_cookbook_content, get_added_content)
get_cookbook_removed_content = functools.partial(_filter_for_cookbook_content, get_removed_content)


def get_content_and_unit_paths(repo):
    """Return the publication relative path of content units present in a cookbook repository.

    :param repo: A dict of information about the repository.
    :returns: A list of tuples (content unit, publication path of the unit) present in a given
              repository.
    """
    # The "relative_path" is actually a file path and name
    def rel_path(content):
        path_format = "cookbook_files/{0}/{2}/{0}-{1}.tar.gz"
        return path_format.format(
            content["name"], content["version"], content["version"].replace(".", "_")
        )

    return [(cu, rel_path(cu)) for cu in get_cookbook_content(repo)]



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from typing import Dict, Iterable, List, Optional
from urllib.parse import urljoin

from pulp_smash.pulp3.constants import BASE_PUBLISHER_PATH, BASE_REMOTE_PATH, CONTENT_PATH

DOWNLOAD_POLICIES = ["immediate", "on_demand", "streamed"]

FIXTURES_BASE_URL = "https://pulpcookbook.s3.wasabisys.com/"

COOKBOOK_CONTENT_NAME = "cookbook.cookbook"
COOKBOOK_CONTENT_PATH = urljoin(CONTENT_PATH, "cookbook/cookbooks/")

COOKBOOK_REMOTE_PATH = urljoin(BASE_REMOTE_PATH, "cookbook/cookbook/")

COOKBOOK_PUBLISHER_PATH = urljoin(BASE_PUBLISHER_PATH, "cookbook/cookbook/")

COOKBOOK_BASE_CONTENT_URL = "http://localhost:8080/pulp_cookbook/content/"


class CookbookFixture:
    """Class holding information about a universe cookbook repo fixture."""

    def __init__(self, repo_name: str, cookbooks: Dict[str, List[str]]) -> None:
        self.url = urljoin(FIXTURES_BASE_URL, repo_name + "/")
        self.cookbooks = cookbooks

    def cookbook_count(self, names: Optional[Iterable] = None) -> int:
        """Return the number of cookbooks or the number of cookbooks with a specific name."""
        if not names:
            names = self.cookbooks.keys()
        return sum(len(self.cookbooks[name]) for name in names)

    def cookbook_download_url(self, name: str, version: str):
        uversion = version.replace(".", "_")
        return urljoin(self.url, f"cookbook_files/{name}/{uversion}/{name}-{version}.tar.gz")


fixture_u1 = CookbookFixture(
    "u1",
    {
        "apache2": ["5.0.1", "5.0.0", "3.0.1", "1.10.2", "1.9.6", "1.9.1"],
        "apt": ["7.0.0", "6.0.1", "5.0.1"],
        "ubuntu": ["2.0.1", "1.0.0"],
        "nginx": ["8.0.0", "2.7.4", "2.7.2", "2.7.0", "2.6.0", "2.5.0"],
    },
)
fixture_u1.example1_name = "ubuntu"
fixture_u1.example1_version = "2.0.1"
fixture_u1.example2_name = "nginx"
fixture_u1.example2_version = "2.7.4"

# u1_diff_digest1 has the same cookbooks as u1, but the tar.gz files are
# different. Used in repo isolation tests together with u1
fixture_u1_diff_digest = CookbookFixture("u1_diff_digest", fixture_u1.cookbooks)


COOKBOOK2_FIXTURE_URL = urljoin(FIXTURES_BASE_URL, "u2/")
"""The URL to a cookbook universe repository."""

COOKBOOK2_URL = urljoin(COOKBOOK2_FIXTURE_URL, "1.iso")
"""The URL to an ISO cookbook at :data:`COOKBOOK2_FIXTURE_URL`."""



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging
import os

from gettext import gettext as _

from django.core.files import File
from django.db.models import Count

from pulpcore.plugin.models import (
    RepositoryVersion,
    Publication,
    PublishedArtifact,
    PublishedMetadata,
)
from pulpcore.plugin.tasking import WorkingDirectory

from pulp_cookbook.app.models import CookbookPackageContent, CookbookPublisher
from pulp_cookbook.metadata import Entry, Universe

log = logging.getLogger(__name__)

BASE_PATH_MARKER = "{{ base_path }}"


def path_template(artifact_path):
    """Artifact path to write into published '__universe__' file."""
    return BASE_PATH_MARKER + "/" + artifact_path


def replace_all_paths(content, base_url):
    """Replace BASE_PATH_MARKERs with base_url in the give '__universe__' content."""
    return content.replace(BASE_PATH_MARKER, base_url)


def publish(publisher_pk, repository_version_pk):
    """
    Use provided publisher to create a Publication based on a RepositoryVersion.

    Args:
        publisher_pk (str): Use the publish settings provided by this publisher.
        repository_version_pk (str): Create a publication from this repository version.
    """
    publisher = CookbookPublisher.objects.get(pk=publisher_pk)
    repository_version = RepositoryVersion.objects.get(pk=repository_version_pk)

    log.info(
        _("Publishing: repository=%(repository)s, version=%(version)d, publisher=%(publisher)s"),
        {
            "repository": repository_version.repository.name,
            "version": repository_version.number,
            "publisher": publisher.name,
        },
    )

    with WorkingDirectory():
        with Publication.create(repository_version, publisher) as publication:
            check_repo_version_constraint(publication)
            universe = Universe("__universe__")
            universe.write(populate(publication))
            metadata = PublishedMetadata(
                relative_path=os.path.basename(universe.relative_path),
                publication=publication,
                file=File(open(universe.relative_path, "rb")),
            )
            metadata.save()

    log.info(_("Publication: %(publication)s created"), {"publication": publication.pk})


def check_repo_version_constraint(publication):
    """
    Ensure that repo version to publish fulfills repo_key_fields() uniqueness.

    Raises:
        ValueError: When constraint is violated

    """
    fields = CookbookPackageContent.repo_key_fields()
    qs_content = CookbookPackageContent.objects.filter(
        pk__in=publication.repository_version.content
    )
    qs = qs_content.values(*fields).annotate(num_cookbooks=Count("pk")).filter(num_cookbooks__gt=1)
    duplicates = [f"{res['name']} {res['version']}" for res in qs]
    if duplicates:
        raise ValueError(
            f"Publication would contain multiple versions of cookbooks: {', '.join(duplicates)}"
        )


def populate(publication):
    """
    Populate a publication.

    Create published artifacts and yield a Universe Entry for each.

    Args:
        publication (pulpcore.plugin.models.Publication): A Publication to populate.

    Yields:
        Entry: Universe entry for each cookbook

    """
    for content in CookbookPackageContent.objects.filter(
        pk__in=publication.repository_version.content
    ).order_by("-_created"):
        relative_path = "cookbook_files/{}/{}/".format(
            content.name, content.version.replace(".", "_")
        )
        for content_artifact in content.contentartifact_set.all():
            art_path = os.path.join(relative_path, content_artifact.relative_path)
            published_artifact = PublishedArtifact(
                relative_path=art_path, publication=publication, content_artifact=content_artifact
            )
            published_artifact.save()
            entry = Entry(
                name=content.name,
                version=content.version,
                download_url=path_template(art_path),
                dependencies=content.dependencies,
            )
            yield entry



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from unittest.mock import Mock

from django.test import TestCase

from pulp_cookbook.app.models import CookbookPackageContent
from pulp_cookbook.app.tasks.publishing import check_repo_version_constraint


class PublishingCheckRepoVersionConstraintTestCase(TestCase):
    """Verify check_repo_version_constraint() method."""

    def new_publication_version_content(self):
        publication = Mock()
        publication.repository_version.content = CookbookPackageContent.objects.all()
        return publication

    def test_check_repo_version_constraint_ok(self):
        CookbookPackageContent.objects.create(
            name="c1", version="1.0.0", content_id_type="sha256", content_id="1", dependencies={}
        )
        CookbookPackageContent.objects.create(
            name="c2", version="1.0.0", content_id_type="sha256", content_id="1", dependencies={}
        )
        CookbookPackageContent.objects.create(
            name="c1", version="1.0.1", content_id_type="sha256", content_id="2", dependencies={}
        )

        check_repo_version_constraint(self.new_publication_version_content())

    def test_check_repo_version_constraint_not_ok(self):
        CookbookPackageContent.objects.create(
            name="c1", version="1.0.0", content_id_type="sha256", content_id="1", dependencies={}
        )
        CookbookPackageContent.objects.create(
            name="c1", version="1.0.0", content_id_type="sha256", content_id="2", dependencies={}
        )
        CookbookPackageContent.objects.create(
            name="c2", version="1.0.0", content_id_type="sha256", content_id="1", dependencies={}
        )
        CookbookPackageContent.objects.create(
            name="c2", version="1.0.0", content_id_type="sha256", content_id="2", dependencies={}
        )
        CookbookPackageContent.objects.create(
            name="c1", version="1.0.1", content_id_type="sha256", content_id="1", dependencies={}
        )
        with self.assertRaises(ValueError):
            check_repo_version_constraint(self.new_publication_version_content())

    def test_check_repo_version_constraint_same_and_different_id_types_not_ok(self):
        CookbookPackageContent.objects.create(
            name="c1", version="1.0.0", content_id_type="sha256", content_id="1", dependencies={}
        )
        CookbookPackageContent.objects.create(name="c1", version="1.0.0", dependencies={})
        CookbookPackageContent.objects.create(
            name="c1", version="1.0.1", content_id_type="sha256", content_id="2", dependencies={}
        )

        with self.assertRaises(ValueError):
            check_repo_version_constraint(self.new_publication_version_content())



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Tests that publish cookbook plugin repositories."""
import unittest
from random import choice
from urllib.parse import urljoin

from requests.exceptions import HTTPError

from pulp_smash import api, config
from pulp_smash.exceptions import TaskReportError
from pulp_smash.pulp3.constants import REPO_PATH
from pulp_smash.pulp3.utils import delete_orphans, gen_remote, gen_repo, get_versions, publish, sync

from pulp_cookbook.tests.functional.api.utils import gen_publisher, get_cookbook_content
from pulp_cookbook.tests.functional.constants import (
    fixture_u1,
    fixture_u1_diff_digest,
    COOKBOOK_REMOTE_PATH,
    COOKBOOK_PUBLISHER_PATH,
)


class PublishAnyRepoVersionTestCase(unittest.TestCase):
    """Test whether a particular repository version can be published.

    This test targets the following issues:

    * `Pulp #3324 <https://pulp.plan.io/issues/3324>`_
    * `Pulp Smash #897 <https://github.com/PulpQE/pulp-smash/issues/897>`_
    """

    def test_all(self):
        """Test whether a particular repository version can be published.

        1. Create a repository with at least 2 repository versions.
        2. Create a publication by supplying the latest ``repository_version``.
        3. Assert that the publication ``repository_version`` attribute points
           to the latest repository version.
        4. Create a publication by supplying the non-latest
           ``repository_version``.
        5. Assert that the publication ``repository_version`` attribute points
           to the supplied repository version.
        6. Assert that an exception is raised when providing two different
           repository versions to be published at same time.
        """
        cfg = config.get_config()

        delete_orphans(cfg)

        client = api.Client(cfg, api.json_handler)
        body = gen_remote(fixture_u1.url, cookbooks={fixture_u1.example1_name: ""})
        remote = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["_href"])

        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["_href"])

        sync(cfg, remote, repo, mirror=True)
        repo = client.get(repo["_href"])
        repo_content = get_cookbook_content(repo)
        self.assertTrue(repo_content)

        publisher = client.post(COOKBOOK_PUBLISHER_PATH, gen_publisher())
        self.addCleanup(client.delete, publisher["_href"])

        # Step 1
        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["_href"])
        for cookbook in repo_content:
            client.post(repo["_versions_href"], {"add_content_units": [cookbook["_href"]]})
        version_hrefs = tuple(ver["_href"] for ver in get_versions(repo))
        non_latest = choice(version_hrefs[:-1])

        # Step 2
        publication = publish(cfg, publisher, repo)

        # Step 3
        self.assertEqual(publication["repository_version"], version_hrefs[-1])

        # Step 4
        publication = publish(cfg, publisher, repo, non_latest)

        # Step 5
        self.assertEqual(publication["repository_version"], non_latest)

        # Step 6
        with self.assertRaises(HTTPError):
            body = {"repository": repo["_href"], "repository_version": non_latest}
            client.post(urljoin(publisher["_href"], "publish/"), body)


class RepoVersionConstraintValidationTestCase(unittest.TestCase):
    """Test the check for repo version constraints."""

    def test_publish_invalid_repo_version(self):
        """Repo version containing two units with the same name and version can't be published."""
        cfg = config.get_config()
        delete_orphans(cfg)
        client = api.Client(cfg, api.json_handler)

        # Create repo u1 and sync partially
        repo_u1 = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo_u1["_href"])

        body = gen_remote(fixture_u1.url, cookbooks={fixture_u1.example1_name: ""})
        remote_u1 = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote_u1["_href"])

        sync(cfg, remote_u1, repo_u1, mirror=True)
        repo_u1 = client.get(repo_u1["_href"])

        # Create repo u1_diff_digest and sync partially
        repo_u1_diff_digest = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo_u1_diff_digest["_href"])

        body = gen_remote(fixture_u1_diff_digest.url, cookbooks={fixture_u1.example1_name: ""})
        remote_u1_diff_digest = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote_u1_diff_digest["_href"])

        sync(cfg, remote_u1_diff_digest, repo_u1_diff_digest, mirror=True)
        repo_u1_diff_digest = client.get(repo_u1_diff_digest["_href"])

        # Add a content unit from u1_diff_digest to u1 (duplicate name&version)
        content_u1_diff_digest = get_cookbook_content(repo_u1_diff_digest)
        self.assertTrue(content_u1_diff_digest)
        client.post(
            repo_u1["_versions_href"], {"add_content_units": [content_u1_diff_digest[0]["_href"]]}
        )

        publisher = client.post(COOKBOOK_PUBLISHER_PATH, gen_publisher())
        self.addCleanup(client.delete, publisher["_href"])
        with self.assertRaisesRegex(TaskReportError, "would contain multiple versions"):
            publish(cfg, publisher, repo_u1)



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import logging

from collections import defaultdict
from gettext import gettext as _
from urllib.parse import urljoin

from django.db.models import Prefetch, Q

from pulpcore.plugin.models import Artifact, ContentArtifact, ProgressBar, Remote, Repository
from pulpcore.plugin.stages import (
    DeclarativeArtifact,
    DeclarativeContent,
    DeclarativeVersion,
    Stage,
    ArtifactDownloader,
    ArtifactSaver,
    RemoteArtifactSaver,
    ContentSaver,
)

from pulp_cookbook.app.models import CookbookPackageContent, CookbookRemote
from pulp_cookbook.metadata import Universe

log = logging.getLogger(__name__)


class UpdateContentWithDownloadResult(Stage):
    """
    A stage that sets the content_id (SHA256 from the artifact).

    Sate the content_id to the SHA256 checksum for "unsaved" content or, if
    saved content has no/a different checksum, create a new content instance
    with a SHA256 content_id.
    """

    async def run(self):
        async for d_content in self.items():
            download_sha256 = d_content.d_artifacts[0].artifact.sha256
            if d_content.content._state.adding:
                d_content.content.set_sha256_digest(download_sha256)
            else:
                if d_content.content.content_id != download_sha256:
                    # To keep previous repository versions untouched, create a
                    # new content unit instead of modifying the existing
                    # content.
                    # To copy multiple inheritance models, we need to set both
                    # pk and _id to None and, as stages look at _state.adding,
                    # reset that as well...
                    d_content.content.pk = None
                    d_content.content._id = None
                    d_content.content._state.adding = True
                    d_content.content.set_sha256_digest(download_sha256)
            await self.put(d_content)


class QueryExistingRepoContentAndArtifacts(Stage):
    """
    A Stages API stage that searches for saved Content and Artifacts in the base repo version.

    This stage expects :class:`~pulpcore.plugin.stages.DeclarativeContent` units
    from `in_q` and inspects their associated
    :class:`~pulpcore.plugin.models.Content` instance.

    This stage inspects any "unsaved" Content unit objects and searches for
    existing saved Content units with the same key. The search is constrained to
    the content of the given repository version `new_version` and use the query
    :class:`~pulpcore.plugin.models.Content.repo_q()` and the key
    :class:`~pulpcore.plugin.models.Content.repo_key_fields()`.

    Any existing Content objects found replace their "unsaved" counterpart in
    the :class:`~pulpcore.plugin.stages.DeclarativeContent` object.
    Additionally, inspect the associated
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` units:  Any existing
    Artifact objects found replace their "unsaved" counterpart in the
    :class:`~pulpcore.plugin.stages.DeclarativeArtifact` object (the match is
    done on `relative_path`).

    Each :class:`~pulpcore.plugin.stages.DeclarativeContent` is sent to `out_q`
    after it has been handled.

    This stage drains all available items from `in_q` and batches everything
    into one large call to the db for efficiency.

    Args: new_version (:class:`~pulpcore.plugin.models.RepositoryVersion`):
        Optional repository version to search content in.

    """

    def __init__(self, new_version, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_version = new_version

    async def run(self):
        async for batch in self.batches():
            self._process_batch(batch)
            for declarative_content in batch:
                await self.put(declarative_content)

    def _process_batch(self, batch):
        unsaved_d_cs = [dc for dc in batch if dc.content._state.adding]

        content_q_by_type = defaultdict(lambda: Q(pk=None))
        # declarative_content by model type and repo key
        d_c_by_mt_rk = defaultdict(dict)

        for declarative_content in unsaved_d_cs:
            m_type = type(declarative_content.content)
            unit_q = declarative_content.content.repo_q()
            content_q_by_type[m_type] = content_q_by_type[m_type] | unit_q
            d_c_by_mt_rk[m_type][declarative_content.content.repo_key()] = declarative_content

        for model_type in content_q_by_type:
            self._associate_model_type(
                model_type, content_q_by_type[model_type], d_c_by_mt_rk[model_type]
            )

    def _associate_model_type(self, model_type, q, d_c_by_repo_key):
        content_filter = model_type.objects.filter(pk__in=self.new_version.content)
        content_filter = content_filter.filter(q)
        # prefetch the related ContentArtifact and Artifact objects:
        content_filter = content_filter.prefetch_related(
            Prefetch(
                "_artifacts",
                queryset=(
                    ContentArtifact.objects.filter(artifact__isnull=False).select_related(
                        "artifact"
                    )
                ),
                to_attr="c_as_with_artifact",
            )
        )
        for content in content_filter:
            repo_key = content.repo_key()
            try:
                declarative_content = d_c_by_repo_key[repo_key]
            except KeyError:
                pass
            else:
                declarative_content.content = content
                for c_a in content.c_as_with_artifact:
                    for d_a in declarative_content.d_artifacts:
                        if d_a.relative_path == c_a.relative_path:
                            d_a.artifact = c_a.artifact


# TODO: search within repo version not needed because of
#       QueryExistingRepoContentAndArtifacts. Is there any use for this stage?
class QueryExistingContentUnits(Stage):
    """
    A Stages API stage that searches for existing saved Content units.

    This stage expects :class:`~pulpcore.plugin.stages.DeclarativeContent` units
    from `in_q` and inspects their associated
    :class:`~pulpcore.plugin.models.Content` instance.

    This stage inspects any "unsaved" Content unit objects and searches for
    existing saved Content units inside Pulp with the same key. The search
    depends in the value of `new_version`:

    - If present, search in the content of the given repository version using
      the key :class:`~pulpcore.plugin.models.Content.repo_key_fields()`.

    - If not present (`None`), search in the content of Pulp using
      the default :class:`~pulpcore.plugin.models.Content.q()` method.

    Any existing Content objects found, replace their "unsaved" counterpart in
    the :class:`~pulpcore.plugin.stages.DeclarativeContent` object.

    Each :class:`~pulpcore.plugin.stages.DeclarativeContent` is sent to `out_q`
    after it has been handled.

    This stage drains all available items from `in_q` and batches everything
    into one large call to the db for efficiency.

    Args:
        new_version (:class:`~pulpcore.plugin.models.RepositoryVersion`): Optional
            repository version to search content in.
    """

    def __init__(self, new_version=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.new_version = new_version

    async def run(self):
        async for batch in self.batches():
            self._process_batch(batch)
            for declarative_content in batch:
                await self.put(declarative_content)

    def _process_batch(self, batch):
        unsaved_d_cs = [dc for dc in batch if dc.content._state.adding]
        content_q_by_type = defaultdict(lambda: Q(pk=None))
        for declarative_content in unsaved_d_cs:
            model_type = type(declarative_content.content)
            if self.new_version:
                unit_q = declarative_content.content.repo_q()
            else:
                unit_q = declarative_content.content.q()
            content_q_by_type[model_type] = content_q_by_type[model_type] | unit_q

        for model_type in content_q_by_type.keys():
            content_filter = model_type.objects.filter(content_q_by_type[model_type])
            if self.new_version:
                content_filter = content_filter.filter(pk__in=self.new_version.content)
                key_fields = model_type.repo_key_fields()
            else:
                key_fields = model_type.natural_key_fields()
            for result in content_filter:
                for declarative_content in unsaved_d_cs:
                    if type(declarative_content.content) is not model_type:
                        continue
                    not_same_unit = False
                    for field in key_fields:
                        in_memory_digest_value = getattr(declarative_content.content, field)
                        if in_memory_digest_value != getattr(result, field):
                            not_same_unit = True
                            break
                    if not_same_unit:
                        continue
                    declarative_content.content = result


class CookbookFirstStage(Stage):
    """The first stage of the pulp_cookbook sync pipeline."""

    def __init__(self, remote, download_artifacts, *args, **kwargs):
        """
        The first stage of the pulp_cookbook sync pipeline.

        Get the `universe` of the remote repo, parse it and inject a
        DeclarativeContent instance for each cookbook found.

        Args:
            remote (CookbookRemote): The remote data to be used when syncing
        """
        super().__init__(*args, **kwargs)
        self.remote = remote
        self.download_artifacts = download_artifacts

    async def run(self):
        """
        Build and emit `DeclarativeContent` from the Manifest data.

        If a cookbook specifier is set in the remote, cookbooks are filtered
        using this specifier.

        """
        with ProgressBar(message="Downloading Metadata", total=1) as pb:
            downloader = self.remote.get_downloader(url=urljoin(self.remote.url + "/", "universe"))
            result = await downloader.run()
            pb.increment()

        cookbook_names = self.remote.specifier_cookbook_names()

        with ProgressBar(message="Parsing Metadata") as pb:
            universe = Universe(result.path)
            for entry in universe.read():
                if cookbook_names and entry.name not in cookbook_names:
                    continue
                cookbook = CookbookPackageContent(
                    name=entry.name, version=entry.version, dependencies=entry.dependencies
                )
                artifact = Artifact()
                da = DeclarativeArtifact(
                    artifact=artifact,
                    url=entry.download_url,
                    relative_path=cookbook.relative_path(),
                    remote=self.remote,
                    deferred_download=not self.download_artifacts,
                )
                dc = DeclarativeContent(content=cookbook, d_artifacts=[da])
                pb.increment()
                await self.put(dc)


class CookbookDeclarativeVersion(DeclarativeVersion):
    """Implement pulp_cookbook's stage API pipeline."""

    def __init__(self, download_artifacts, *args, **kwargs):
        self.download_artifacts = download_artifacts
        super().__init__(*args, **kwargs)

    def pipeline_stages(self, new_version):
        pipeline = [self.first_stage, QueryExistingRepoContentAndArtifacts(new_version=new_version)]
        if self.download_artifacts:
            pipeline.extend(
                [
                    ArtifactDownloader(),
                    ArtifactSaver(),
                    UpdateContentWithDownloadResult(),
                    QueryExistingContentUnits(),  # share content with known digest
                ]
            )
        pipeline.append(ContentSaver())
        pipeline.append(RemoteArtifactSaver())
        return pipeline


def synchronize(remote_pk, repository_pk, mirror):
    """
    Create a new version of the repository that is synchronized with the remote.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.

    Raises:
        ValueError: When url is empty.

    """
    remote = CookbookRemote.objects.get(pk=remote_pk)
    repository = Repository.objects.get(pk=repository_pk)
    if not remote.url:
        raise ValueError(_("A remote must have a url specified to synchronize."))

    download = remote.policy == Remote.IMMEDIATE  # Interpret policy to download Artifacts or not

    first_stage = CookbookFirstStage(remote=remote, download_artifacts=download)
    dv = CookbookDeclarativeVersion(
        first_stage=first_stage, repository=repository, mirror=mirror, download_artifacts=download
    )
    dv.create()



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

from unittest.mock import Mock

from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase

from pulpcore.plugin.models import Artifact, ContentArtifact
from pulpcore.plugin.stages import DeclarativeArtifact, DeclarativeContent

from pulp_cookbook.app.models import CookbookPackageContent, CookbookRemote
from pulp_cookbook.app.tasks.synchronizing import (
    QueryExistingContentUnits,
    QueryExistingRepoContentAndArtifacts,
)


class QueryExistingContentUnitsTestCase(TestCase):
    """Verify QueryExistingContentUnits stage for both natural and repo keys."""

    def setUp(self):
        self.c1 = CookbookPackageContent.objects.create(
            name="c1", version="1.0.0", content_id="1", dependencies={}
        )

    def new_version_all_content(self):
        """Provide a mock for a repository version containing all content."""
        new_version = Mock()
        new_version.content = CookbookPackageContent.objects.all()
        return new_version

    def test_content_associated_using_repo_key(self):
        dc_c1 = CookbookPackageContent(name="c1", version="1.0.0", dependencies={})
        dc_c2 = CookbookPackageContent(name="c2", version="1.0.0", dependencies={})

        batch = [
            DeclarativeContent(content=dc_c1, d_artifacts=[]),
            DeclarativeContent(content=dc_c2, d_artifacts=[]),
        ]
        stage = QueryExistingContentUnits(new_version=self.new_version_all_content())
        stage._process_batch(batch)
        self.assertEqual(batch[0].content.content_id, "1")
        self.assertEqual(batch[0].content.pk, self.c1.pk)
        self.assertIsNone(batch[1].content.pk)

    def test_content_associated_using_natural_key(self):
        dc_c1 = CookbookPackageContent(name="c1", version="1.0.0", content_id="1", dependencies={})
        dc_c1_other = CookbookPackageContent(
            name="c1", version="1.0.0", content_id="other", dependencies={}
        )

        batch = [
            DeclarativeContent(content=dc_c1, d_artifacts=[]),
            DeclarativeContent(content=dc_c1_other, d_artifacts=[]),
        ]
        stage = QueryExistingContentUnits()
        stage._process_batch(batch)
        self.assertEqual(batch[0].content.content_id, "1")
        self.assertEqual(batch[0].content.pk, self.c1.pk)
        self.assertIsNone(batch[1].content.pk)

    def test_existing_content_is_ok(self):
        # dc_c1 is a duplicate of c1 existing in the DB
        dc_c1 = CookbookPackageContent(name="c1", version="1.0.0", content_id="1", dependencies={})
        c2 = CookbookPackageContent.objects.create(
            name="c2", version="1.0.0", content_id="2", dependencies={}
        )

        batch = [
            DeclarativeContent(content=dc_c1, d_artifacts=[]),
            DeclarativeContent(content=c2, d_artifacts=[]),
        ]
        stage = QueryExistingContentUnits()
        stage._process_batch(batch)
        self.assertEqual(batch[0].content.pk, self.c1.pk)
        self.assertEqual(batch[1].content.pk, c2.pk)


class QueryExistingRepoContentAndArtifactsTestCase(TestCase):
    """Verify that the QueryExistingRepoContentAndArtifacts properly associates from DB."""

    def setUp(self):
        self.remote = CookbookRemote.objects.create(name="remote")
        # c1: Existing content unit with Artifact a1
        self.a1 = Artifact.objects.create(
            size=0,
            sha256="1",
            sha384="1",
            sha512="1",
            file=SimpleUploadedFile("test_filename", b""),
        )
        self.c1 = CookbookPackageContent.objects.create(
            name="c1",
            version="1.0.0",
            content_id_type=CookbookPackageContent.SHA256,
            content_id="1",
            dependencies={},
        )
        ContentArtifact.objects.create(
            artifact=self.a1, content=self.c1, relative_path=self.c1.relative_path()
        )
        # c3: content unit does exist, has a content_artifact association,
        # but no artifact
        self.c3 = CookbookPackageContent.objects.create(name="c3", version="1.0.0", dependencies={})
        ContentArtifact.objects.create(
            artifact=None, content=self.c3, relative_path=self.c3.relative_path()
        )

    def new_version_all_content(self):
        """Provide a mock for a repository version containing all content."""
        new_version = Mock()
        new_version.content = CookbookPackageContent.objects.all()
        return new_version

    def test_content_associated_using_repo_key(self):
        stage = QueryExistingRepoContentAndArtifacts(new_version=self.new_version_all_content())

        # c1: Existing content unit with Artifact
        c1 = CookbookPackageContent(name="c1", version="1.0.0", dependencies={})
        # c2: content unit does not exist in DB
        c2 = CookbookPackageContent(name="c2", version="1.0.0", dependencies={})
        # c3: content unit does exist, has a content_artifact association,
        # but no artifact (i.e. is a non-immediate content unit)
        c3 = CookbookPackageContent(name="c3", version="1.0.0", dependencies={})

        d_c1_d_a1 = DeclarativeArtifact(
            artifact=Artifact(),
            url="http://a1",
            relative_path=c1.relative_path(),
            remote=self.remote,
        )
        d_c2_d_a2 = DeclarativeArtifact(
            artifact=Artifact(),
            url="http://a2",
            relative_path=c2.relative_path(),
            remote=self.remote,
        )
        d_c3_d_a3 = DeclarativeArtifact(
            artifact=Artifact(),
            url="http://a3",
            relative_path=c3.relative_path(),
            remote=self.remote,
        )

        batch = [
            DeclarativeContent(content=c1, d_artifacts=[d_c1_d_a1]),
            DeclarativeContent(content=c2, d_artifacts=[d_c2_d_a2]),
            DeclarativeContent(content=c3, d_artifacts=[d_c3_d_a3]),
        ]

        stage._process_batch(batch)

        self.assertEqual(batch[0].content.content_id, "1")
        self.assertEqual(batch[0].content.pk, self.c1.pk)
        self.assertEqual(batch[0].d_artifacts[0].artifact.pk, self.a1.pk)

        self.assertIsNone(batch[1].content.pk)
        self.assertTrue(batch[1].d_artifacts[0].artifact._state.adding)

        self.assertEqual(batch[2].content.pk, self.c3.pk)
        self.assertTrue(batch[2].d_artifacts[0].artifact._state.adding)



# (C) Copyright 2019 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

"""Tests that verify download of content served by Pulp."""
import hashlib
import unittest
from random import choice
from urllib.parse import urljoin

from pulp_smash import api, config, exceptions, utils
from pulp_smash.pulp3.constants import DISTRIBUTION_PATH, REPO_PATH
from pulp_smash.pulp3.utils import (
    delete_orphans,
    gen_distribution,
    gen_remote,
    gen_repo,
    publish,
    sync,
)

from pulp_cookbook.tests.functional.api.utils import (
    gen_publisher,
    get_content_and_unit_paths,
    get_cookbook_content,
)
from pulp_cookbook.tests.functional.constants import (
    fixture_u1,
    COOKBOOK_REMOTE_PATH,
    COOKBOOK_PUBLISHER_PATH,
    COOKBOOK_BASE_CONTENT_URL,
)


class DownloadContentTestCase(unittest.TestCase):
    """Verify whether content served by pulp can be downloaded."""

    def create_and_sync_repo(self, cfg, client, policy):
        repo = client.post(REPO_PATH, gen_repo())
        self.addCleanup(client.delete, repo["_href"])

        body = gen_remote(fixture_u1.url, policy=policy)
        remote = client.post(COOKBOOK_REMOTE_PATH, body)
        self.addCleanup(client.delete, remote["_href"])

        sync(cfg, remote, repo, mirror=True)
        return client.get(repo["_href"])

    def create_distribution(self, cfg, client, repo):
        """Create a publication for the latest repo version and a distribution."""
        # Create a publisher.
        publisher = client.post(COOKBOOK_PUBLISHER_PATH, gen_publisher())
        self.addCleanup(client.delete, publisher["_href"])

        # Create a publication.
        publication = publish(cfg, publisher, repo)
        self.addCleanup(client.delete, publication["_href"])

        # Create a distribution.
        body = gen_distribution()
        body["publication"] = publication["_href"]
        response_dict = client.post(DISTRIBUTION_PATH, body)
        dist_task = client.get(response_dict["task"])
        distribution_href = dist_task["created_resources"][0]
        distribution = client.get(distribution_href)
        self.addCleanup(client.delete, distribution_href)
        return distribution

    def download_check(self, cfg, client, repo, distribution, policy):
        """Verify the '/universe' endpoint, download and check a cookbook."""
        # pulp_cookbook universe live endpoint contains
        # all cookbooks
        distribution_base_url = cfg.get_hosts("api")[0].roles["api"]["scheme"]
        distribution_base_url += "://" + distribution["base_url"] + "/"

        # TODO: HACK: can't set the base_url correctly currently
        distribution_base_url = distribution_base_url.replace("/pulp/", "/pulp_cookbook/")

        universe_url = COOKBOOK_BASE_CONTENT_URL
        universe_url += distribution["base_path"] + "/universe"

        universe = client.get(universe_url)
        for content, publication_path in get_content_and_unit_paths(repo):
            u_url = universe[content["name"]][content["version"]]["download_url"]
            self.assertEqual(u_url, urljoin(distribution_base_url, publication_path))

        # Pick a cookbook, and download it from Fixtures…
        cu, unit_path = choice(get_content_and_unit_paths(repo))
        fixtures_hash = hashlib.sha256(
            utils.http_get(urljoin(fixture_u1.url, unit_path))
        ).hexdigest()

        # …and Pulp content
        client.response_handler = api.safe_handler

        unit_url = urljoin(distribution_base_url, unit_path)

        pulp_hash = hashlib.sha256(client.get(unit_url).content).hexdigest()
        self.assertEqual(fixtures_hash, pulp_hash)

        # Verify that the content unit contains the right sha256
        cu_updated = client.get(cu["_href"]).json()
        if policy != "streamed":
            self.assertEqual(fixtures_hash, cu_updated["content_id"])
            self.assertIsNotNone(cu_updated["_artifact"])
        else:
            self.assertIsNone(cu_updated["_artifact"])

    def sync_and_download_check(self, policy):
        """Verify whether content served by pulp can be downloaded.

        The process of publishing content is more involved in Pulp 3 than it
        was under Pulp 2. Given a repository, the process is as follows:

        1. Create a publication from the repository. (The latest repository
           version is selected if no version is specified.) A publication is a
           repository version plus metadata.
        2. Create a distribution from the publication. The distribution defines
           at which URLs a publication is available. For the cookbook plugin
           this is below a live API at ``pulp_cookbook/market/`` e.g.
           ``http://example.com/pulp_cookbook/market/foo/`` and
           ``http://example.com/pulp_cookbook/market/bar/``.

        Do the following:

        1. Create, populate, publish, and distribute a repository.
        2. Select a random content unit in the distribution. Download that
           content unit from Pulp, and verify that the content unit has the
           same checksum when fetched directly from fixtures.

        This test targets the following issues:

        * `Pulp #2895 <https://pulp.plan.io/issues/2895>`_
        * `Pulp Smash #872 <https://github.com/PulpQE/pulp-smash/issues/872>`_
        """
        cfg = config.get_config()
        delete_orphans(cfg)
        client = api.Client(cfg, api.json_handler)
        repo = self.create_and_sync_repo(cfg, client, policy)
        distribution = self.create_distribution(cfg, client, repo)
        self.download_check(cfg, client, repo, distribution, policy)

    def sync_merge_and_download_check(self, policy):
        """
        Merge two repos synced from same source and publish.

        1. sync repo and repo2 from the same source
        2. Add all content units from repo2 to repo
        3. publish and create distribution for repo
        4. Perform download check
        """
        cfg = config.get_config()
        delete_orphans(cfg)
        client = api.Client(cfg, api.json_handler)
        repo = self.create_and_sync_repo(cfg, client, policy)
        repo2 = self.create_and_sync_repo(cfg, client, policy)

        cb_content = get_cookbook_content(repo2)
        client.post(
            repo["_versions_href"], {"add_content_units": [cb["_href"] for cb in cb_content]}
        )
        repo = client.get(repo["_href"])
        distribution = self.create_distribution(cfg, client, repo)
        self.download_check(cfg, client, repo, distribution, policy)

    def test_download_immediate_policy(self):
        self.sync_and_download_check(policy="immediate")

    def test_download_on_demand_policy(self):
        self.sync_and_download_check(policy="on_demand")

    def test_download_streamed_policy(self):
        self.sync_and_download_check(policy="streamed")

    def test_download_merged_immediate_policy(self):
        # When using immediate policy, same content units are merged (sha256 is
        # known). Therefore, there are no duplicate content units in the repo.
        self.sync_merge_and_download_check(policy="immediate")

    def test_download_merged_on_demand_policy(self):
        #  When using non-immediate policy, content units can't be merged
        #  (sha256 is unknown). Therefore, the publish task must fail.
        with self.assertRaisesRegex(
            exceptions.TaskReportError, r"Publication would contain multiple versions of cookbooks:"
        ):
            self.sync_merge_and_download_check(policy="on_demand")



#!/usr/bin/env python3

from setuptools import setup, find_packages

with open("README.rst") as f:
    long_description = f.read()

setup(
    name="pulp-cookbook",
    version="0.0.4b2",
    description="Cookbook plugin for the Pulp Project",
    long_description=long_description,
    author="Simon Baatz",
    author_email="gmbnomis@gmail.com",
    url="https://github.com/gmbnomis/pulp_cookbook/",
    install_requires=["pulpcore-plugin~=0.1rc1"],
    extras_require={"postgres": ["pulpcore[postgres]"], "mysql": ["pulpcore[mysql]"]},
    include_package_data=True,
    packages=find_packages(exclude=["test"]),
    classifiers=(
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ),
    entry_points={"pulpcore.plugin": ["pulp_cookbook = pulp_cookbook:default_app_config"]},
)



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import aiohttp
import logging

from django.conf import settings
from django.db import IntegrityError, transaction

from pulpcore.plugin.content import Handler, PathNotResolved

from pulp_cookbook.app.tasks.publishing import replace_all_paths

log = logging.getLogger(__name__)


def pulp_cookbook_content_path():
    """Get base cotent path from configuration."""
    components = settings.CONTENT_PATH_PREFIX.split("/")
    components[1] = "pulp_cookbook"
    return "/".join(components)


class CookbookContentHandler(Handler):
    """
    Serve cookbook repositories using the universe API.

    This content handler is pretty standard with two exceptions:
    1. Serve the `/universe` endpoint using dynamic replacement for the URLs
    2. When saving ContentArtifact, try to 'update' the CookbookPackageContent
       with the digest of the downloaded artifact.
    """

    async def handle_universe(self, request):
        """
        Serve the '__universe__' metadata object at '/univere'.

        Since the '/universe' endpoint contains absolute URLs, translate
        the content of '__universe__' on the fly.
        """
        path = request.match_info["path"] + "/universe"
        distribution = Handler._match_distribution(path)
        Handler._permit(request, distribution)
        publication = distribution.publication
        if not publication:
            raise PathNotResolved(path)
        pm = publication.published_metadata.get(relative_path="__universe__")
        try:
            with pm.file.open("rb") as u:
                content = u.read().decode("utf-8")
        except FileNotFoundError:
            return PathNotResolved(path)
        except PermissionError:
            return aiohttp.web_exceptions.HTTPForbidden
        content = replace_all_paths(
            content, self._get_content_base_url(request) + distribution.base_path
        )
        return aiohttp.web.Response(body=content, content_type="application/json", charset="utf-8")

    def _save_artifact(self, download_result, remote_artifact):
        new_artifact = super()._save_artifact(
            download_result=download_result, remote_artifact=remote_artifact
        )
        content_artifact = remote_artifact.content_artifact
        content = content_artifact.content.cast()
        try:
            # Try to 'upgrade' the content unit as the digest is known now. This
            # might fail if another identical content unit exists already.
            # We can just ignore this case: As the artifact will be created nevertheless,
            # this method won't be called again for the content_artifact.
            with transaction.atomic():
                content.set_sha256_digest(new_artifact.sha256)
                content.save()
        except IntegrityError:
            pass
        return new_artifact

    def _get_content_base_url(self, request):
        if settings.CONTENT_HOST:
            host = settings.CONTENT_HOST
        else:
            host = request.host
        host = "{}://{}".format(request.scheme, host)

        return host + pulp_cookbook_content_path()



# (C) Copyright 2018 Simon Baatz <gmbnomis@gmail.com>
#
# SPDX-License-Identifier: GPL-2.0-or-later

import uuid

from django.db import models

from pulpcore.plugin.models import Content, Publisher, Remote
from pulpcore.plugin.fields import JSONField


class CookbookPackageContent(Content):
    """
    The "cookbook" content type.

    Content of this type represents a Chef cookbook uniquely
    identified by name, version and content id.

    Fields:
        name (str): The name of the cookbook.
        version (str): The version of the cookbook.
        dependencies (JSON): The dependency information of the cookbook.
        content_id_type ('uuid' or 'sha256'): The type of the content_id field.
            'uuid' if the corresponding artifact is not known (yet). 'sha256'
            otherwise (the SHA256 digest of the corresponding artifact).
        content_id (str): The content id (random UUID or SHA256 digest of artifact)
    """

    # Constants for the choices of content_id types
    UUID = "uuid"
    SHA256 = "sha256"
    CONTENT_ID_TYPE_CHOICES = (
        (UUID, "On demand/streaming content: arbitrary UUID"),
        (SHA256, "Immediate download content: digest of associated artifact"),
    )

    TYPE = "cookbook"

    name = models.TextField(blank=False, null=False)
    version = models.TextField(blank=False, null=False)
    dependencies = JSONField(blank=False, null=False)
    content_id_type = models.CharField(
        max_length=10, choices=CONTENT_ID_TYPE_CHOICES, blank=False, null=False, default=UUID
    )
    content_id = models.CharField(max_length=64, null=False, blank=False, default=uuid.uuid4)

    def relative_path(self):
        return "{}-{}.tar.gz".format(self.name, self.version)

    @staticmethod
    def relative_path_from_data(data):
        return "{}-{}.tar.gz".format(data["name"], data["version"])

    def set_sha256_digest(self, sha256_digest):
        self.content_id_type = self.SHA256
        self.content_id = sha256_digest

    @classmethod
    def repo_key_fields(self):
        """
        Returns a tuple of the key fields that must be unique within a repository version.

        Returns:
            tuple: The repo key field names.

        """
        return ("name", "version")

    def repo_key(self):
        """
        Get the model's repo key based on repo_key_fields().

        Returns:
            tuple: The repo key.

        """
        return tuple(getattr(self, f) for f in self.repo_key_fields())

    def repo_key_dict(self):
        """
        Get the model's repo key as a dictionary of keys and values.

        Returns:
            dict: The repo key.

        """
        return {key: getattr(self, key) for key in self.repo_key_fields()}

    def repo_q(self):
        """
        Returns a Q object that represents the model within a repository version.
        """
        if not self._state.adding:
            return models.Q(pk=self.pk)
        return models.Q(**self.repo_key_dict())

    class Meta:
        unique_together = ("name", "version", "content_id_type", "content_id")


class CookbookRemote(Remote):
    """
    Remote for "cookbook" content.
    """

    TYPE = "cookbook"

    cookbooks = JSONField(blank=True)

    def specifier_cookbook_names(self):
        if self.cookbooks == "":  # blank JSON field
            return None
        else:
            return set(self.cookbooks.keys())


class CookbookPublisher(Publisher):
    """
    Publisher for "cookbook" content.
    """

    TYPE = "cookbook"
